# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation files
#  (the "Software"), to deal in the Software without restriction,
#  including without limitation the rights to use, copy, modify, merge,
#  publish, distribute, sublicense, and/or sell copies of the Software,
#  and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

from __future__ import with_statement, print_function, division

__authors__ = ["Florian Plaswig"]
__license__ = "MIT"
__copyright__ = "ESRF"
__date__ = "10/10/2019"

import logging
logger = logging.getLogger(__name__)
import numpy
from .fabioimage import FabioImage
from .compression import agi_bitfield


class EsperantoImage(FabioImage):
    """FabIO image class for Esperanto image files
    """

    DESCRIPTION = "CrysAlis Pro Esperanto file format "

    DEFAULT_EXTENSIONS = [".eseperanto", ".esper"]

    HEADER_SEPARATOR = "\x0d\x0a"

    HEADER_KEYS = ["IMAGE",
                   "SPECIAL_CCD_1",
                   "SPECIAL_CCD_2",
                   "SPECIAL_CCD_3",
                   "SPECIAL_CCD_4",
                   "SPECIAL_CCD_5",
                   "TIME",
                   "MONITOR",
                   "AUTORUN",
                   "PIXELSIZE",
                   "TIMESTAMP",
                   "GRIDPATTERN",
                   "STARTANGLESINDEG",
                   "ENDANGLESINDEG",
                   "GONIOMODEL_1",
                   "GONIOMODEL_2",
                   "WAVELENGHTH",
                   "MONOCHROMATOR",
                   "HISTORY",
                   "ESPERANTO_FORMAT",
                   "WAVELENGTH",
                   "ABSTORUN"]

    def __init__(self, *arg, **kwargs):
        """
        Generic constructor
        """
        FabioImage.__init__(self, *arg, **kwargs)
        self.format = ""

    def _readheader(self, infile):
        """
        Read and decode the header of an image:

        :param infile: Opened python file (can be stringIO or bipped file)
        """
        self.header = self.check_header()

        # read the first line of the header
        top_line = infile.read(256).decode('ascii')

        if not top_line[-2:] == self.HEADER_SEPARATOR:
            raise RuntimeError("Unable to read esperanto header: Invalid format of first line")

        try:
            header_line_count = int(top_line.split(' ')[5])
        except Exception as err:
            raise RuntimeError("Unable to determine header size: %s" % err)

        self.header["ESPERANTO_FORMAT"] = ' '.join(top_line.split(' ')[:2])

        # read the remaining lines
        for line_num in range(1, header_line_count):
            line = infile.read(256).decode('ascii')

            if not line[-2:] == self.HEADER_SEPARATOR:
                raise RuntimeError("Unable to read esperanto header: Ivalid format of line %d." % (line_num + 1))

            line = line.rstrip()
            split_line = line.split(' ')
            key = split_line[0]
            if key.rstrip() == '':
                continue

            # if key not in self.HEADER_KEYS:
            #     raise RuntimeError("Unable to read esperanto header: Invalid Key %s in line %d." % (key, line_num))

            self.header[key] = ' '.join(split_line[1:])

        # extract necessary data
        if "IMAGE" not in self.header:
            raise RuntimeError("No IMAGE entry found in header.")

        try:
            image_data = self.header["IMAGE"].split(' ')

            width = int(image_data[0])
            height = int(image_data[1])
        except Exception as err:
            raise RuntimeError("Unable to determine dimensions of file: %s" % err)

        if 256 > width > 4096 or width % 4 != 0 or 256 > height > 4096 or height % 4 != 0:
            logger.warning("The dimensions of the image is (%i, %i) but they should only be between 256 and 4096 and a multiple of 4. This might cause compatibility issues.")

        self.shape = (height, width)

        self.format = self.header["IMAGE"].split(' ')[4][1:-1]
        self._dtype = "int32"

    def read(self, fname, frame=None):
        """
        Try to read image

        :param fname: name of the file
        :param frame: number of the frame
        """

        self.resetvals()
        with self._open(fname) as infile:
            self._readheader(infile)

            if self.format == "4BYTE_LONG":
                try:
                    pixelsize = 4
                    pixelcount = self.shape[0] * self.shape[1]
                    data = numpy.frombuffer(infile.read(pixelsize * pixelcount), dtype=self._dtype).copy()
                    self.data = numpy.reshape(data, self.shape)
                except Exception as err:
                    raise RuntimeError("Exception while reading pixel data %s." % err)
            elif self.format == "AGI_BITFIELD":
                self.raw_data = infile.read()
                logger.warning("AGI_BITFIELD decompression is known to be apporximative ... use those data with caution !")
                try:
                    data = agi_bitfield.decompress(self.raw_data, self.shape)
                except Exception as err:
                    raise RuntimeError("Exception while decompressing pixel data %s." % err)
            else:
                raise RuntimeError("Format not supported %s." % self.format)

        return self

    def _formatheaderline(self, line):
        assert len(line) <= 254
        line += ' ' * (254 - len(line))
        line += self.HEADER_SEPARATOR
        return line.encode("ASCII")

    def write(self, fname):
        """
        Write an image

        :param fname: name of the file
        """

        # create header
        if "IMAGE" not in self.header: # create image entry
            dtype_info = '"%s"' % self.format
            dx, dy = self.shape

            self.header["IMAGE"] = "%d %d 1 1 %s" % (dx, dy, dtype_info)

        else: # update image entry
            dtype_info = '"4BYTE_LONG"'
            dx, dy = self.shape
            split = self.header["IMAGE"].split(' ')

            split[0] = str(dy)
            split[1] = str(dx)
            split[4] = str(dtype_info)

            self.header["IMAGE"] = ' '.join(split)

        if 256 > dx > 4096 or dx % 4 != 0 or 256 > dy > 4096 or dy % 4 != 0:
            logger.warning("The dimensions of the image is (%i, %i) but they should only be between 256 and 4096 and a multiple of 4. This might cause compatibility issues.")

        if "ESPERANTO_FORMAT" not in self.header: # default format
            self.header["ESPERANTO_FORMAT"] = "ESPERANTO FORMAT 1.1"

        self.header = dict(filter(self.header, lambda elem: elem[0] in self.HEADER_KEYS))

        header_top = self._formatheaderline("%s CONSISTING OF %d LINES OF 256 BYTE EACH" %
                                                (self.header["ESPERANTO_FORMAT"], len(self.header)))
        HEADER = header_top
        for header_key in self.header:
            if header_key == "ESPERANTO_FORMAT": #or header_key not in self.HEADER_KEYS:
                continue
            header_val = self.header[header_key]
            HEADER += self._formatheaderline(header_key + ' ' + header_val)


        with self._open(fname, "wb") as outfile:
            outfile.write(HEADER)
            if self.format == "4BYTE_LONG":
                outfile.write(self.data.tobytes())
            elif self.format == "AGI_BITFIELD":
                outfile.write(agi_bitfield.compress(self.data))
            else:
                raise RuntimeError("Format not supported %s." % self.format)

# This is not compatibility with old code:
esperantoimage = EsperantoImage



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
# Unit tests
Reading edf files as originally specified by expg with fabio.module edfimage.py

07/06/2019 PB
"""
from __future__ import print_function, with_statement, division, absolute_import
import unittest
import os
import numpy
import shutil
import io
import logging

logger = logging.getLogger(__name__)

import fabio
from fabio.edfimage import edfimage
from ..utilstest import UtilsTest

#logging.basicConfig(level=logging.DEBUG)
#logging.basicConfig(level=logging.WARNING)
#logging.basicConfig(level=logging.INFO)
#logging.debug('This will get logged')
#logging.warning('This will get logged')
#logging.info('This will get logged')

#==========================================
def fopen(filename=None,frameno=None):
    header=None
    data=None
    frame=None
    logging.debug("fopen(filename={},frameno={})".format(filename,frameno))
    if filename:
        image=fabio.open(filename)
        if frameno is None:
            frameno=0

        nframes=image.nframes

        #if image.classname == "EdfImage":
        if hasattr(image,"npsdframes"):
            npsdframes=image.npsdframes
            nerrorframes=image.nerrorframes
        else:
            npsdframes=nframes
            nerrorframes=0

        if frameno >= npsdframes:
            logging.warning("Psd frame {} out of range: 0 <= {} < {}".format(frameno,frameno,npsdframes))

        if frameno <0:
            if -frameno > nerrorframes:
                logging.warning("Error frame {} out of range: {} <= {} < 0 ".format(frameno,-nerrorframes,frameno))
            frameno += nframes

        if frameno in range(nframes):
            if nframes>1:
                frame = image.getframe(frameno)
                data = frame.data
                header=frame.header
            else:
                # Single frame
                data = image.data
                header= image.header
            frame=fabio.fabioimage.FabioFrame(data=data, header=header)
        else:
            raise IOError("fopen: Cannot access frame: {} (0<=frame<{})".format(frameno, nframes))

    return(frame)

def get_data_counts(shape=None):
    '''
    Counts all items specified by shape
    '''
    if shape is None:
      shape=()
    counts=1
    for ishape in range(0,len(shape)):
        counts*=shape[ishape]
    return(counts)

#============================================

# Hint: Must be defined outside the test case class, otherwise used as test
def test_00(cls,filename,avglist=None,keylist=None):
    """
    Checks the correct shape of the data arrays.
    Compares the mean value of each data frame array, with the number given in avglist.
    filename: name of file to test
    avglist: list of average values for each frame, stops, if not equal
    keylist: list of keys to read, stops, if not available
    If avglist or keylist is shorter than frameno, the last value in the list is used.
    """
    image=fabio.open(filename)

    nframes=image.nframes

    if hasattr(image,"npsdframes"):
        npsdframes=image.npsdframes
        nerrorframes=image.nerrorframes
    else:
        npsdframes=image.nframes
        nerrorframes=0

    # To avoid warnings make different loops over psd data and error data
    #psd data
    for frameno in range(0,npsdframes):
        frame=fopen(filename,frameno)

        # check data shape
        counts=get_data_counts(frame.shape)
        data_counts=get_data_counts(frame.data.shape)

        cls.assertEqual(counts, data_counts,"A:filename={},frameno={}: inconsistent data shapes: header.{},data.{}".
           format(filename,frameno,frame.shape,frame.data.shape))

        # calculate mean value
        sum=numpy.sum(frame.data)
        fmean=sum/counts

        logging.debug("filename={},frameno={},sum={},counts={},fmean={}".format(filename,frameno,sum,counts,fmean))

        # read known mean value from avglist
        if avglist is not None:
           if len(avglist)>frameno:
               avg=avglist[frameno]
           else:
               avg=avglist[-1]
           cls.assertLessEqual(abs(fmean-avg),abs(fmean+avg)*5e-6,"B:filename={},frameno={}: unexpected average value: calculated {}, expected {}".
               format(filename,frameno,fmean,avg))

        # read a key to read from keylist
        if keylist is not None:
            if len(keylist)>frameno:
                key=keylist[frameno]
            else:
                key=keylist[-1]

            if key in frame.header:
               logging.debug("filename={}, frameno={}: '{}' = {}".format(filename,frameno,key,frame.header[key]))
            else:
               logging.debug("filename={}, frameno={}: '{}' = None".format(filename,frameno,key))

            cls.assertIn(key,frame.header,"C:filename={},frameno={}: Missing expected header key '{}'".format(filename,frameno,key))

    #error data
    for frameno in range(0,nerrorframes):
        frame=fopen(filename,-frameno-1)

        # check data shape
        counts=get_data_counts(frame.shape)
        data_counts=get_data_counts(frame.data.shape)

        cls.assertEqual(counts, data_counts,"D:filename={},frameno={}: inconsistent data shapes: header.{},data.{}".
           format(filename,frameno,frame.shape,frame.data.shape))

        # calculate mean value
        sum=numpy.sum(frame.data)
        fmean=sum/counts

        logging.debug("filename={},frameno={},sum={},counts={},fmean={}".format(filename,frameno,sum,counts,fmean))

        # read known mean value from avglist
        if avglist is not None:
           #error frames are taken from the end
           if len(avglist)>nframes-frameno-1:
               avg=avglist[nframes-frameno-1]
           else:
               avg=avglist[-1]
           cls.assertLessEqual(abs(fmean-avg),abs(fmean+avg)*5e-6,"E:filename={},frameno={}: unexpected average value: calculated {}, expected {}".
               format(filename,frameno,fmean,avg))

        # read a key to read from keylist
        if keylist is not None:
            if len(keylist)>nframes-frameno-1:
                key=keylist[nframes-frameno-1]
            else:
                key=keylist[-1]

            if key in frame.header:
               logging.debug("filename={},frameno={}: key='{}'".format(filename,frameno,key,frame.header[key]))
            else:
               logging.debug("filename={},frameno={}: key=None".format(filename,frameno,key))

            cls.assertIn(key,frame.header,"F:filename={},frameno={}: Missing expected header key '{}'".format(filename,frameno,key))


class EdfBlockBoundaryCases(unittest.TestCase):
    """
    Test some special cases
    """

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.files = UtilsTest.resources.getdir("ehf_images2.tar.bz2")
        self.root = os.path.join(self.files[0], "..")

    def test_edfblocktypes(self):
        """
        Test reading (gzipped) extended edf multi frame data files with and without a general block
        The files have been prepared with make_multiedfs.sh.
        The pixel values of each frame are equal to the frame number for checking reading the correct data.
        multi5.edf.gz:      5 frames without general block, preset frameno
        multi5+gblk.edf.gz: 5 frames starting with general block, preset to frameno,
                            DefaultKey added to the general block, it must be available for all frames
        multi5_gzip.edf      : like multi5.edf.gz, but with internal compression
        multi5+gblk_gzip.edf : like multi5+gblk.edf.gz, but with internal compression
        """

        avglist=[0.,1.,2.,3.,4.]
        filename = os.path.join(self.root,"00_edfblocktypes/multi5.edf.gz")
        test_00(self,filename,avglist)
        keylist=["DefaultKey"] # the defaultkey is coming from the general block
        filename = os.path.join(self.root,"00_edfblocktypes/multi5+gblk.edf.gz")
        test_00(self,filename,avglist,keylist)
        filename = os.path.join(self.root,"00_edfblocktypes/multi5+gblk_gzip.edf")
        test_00(self,filename,avglist)
        filename = os.path.join(self.root,"00_edfblocktypes/multi5_gzip.edf")
        test_00(self,filename,avglist)

    def test_edfsingle_raw_bf_gblk(self):
        """
        Reading an uncompressed edf-single-frame-file with references to data
        in externally saved dataframes.
        Check average data value, and verify that the key ExperimentInfo is
        correctly read from the general block.

        """
        avglist=[25743.2]
        filename = os.path.join(self.root,"01_single_raw_bf_gblk/pj19_frelon_00028_raw.ehf")
        keylist=["ExperimentInfo"]
        test_00(self,filename,avglist,keylist)

    def test_edfmulti_raw_bf_gblk(self):
        """
        Reading an uncompressed edf-multi-frame-file with references to data
        in externally saved dataframes and checking the average of each frame.
        Some header values are modified for checking the robustness.
        frames 0..16: Test reading from a multi frame edf file with links to
                      external binary files (ehf)
        frame 17:     Test reading without EDF_BinaryFileSize => must be
                      estimated from data size
        frame 18:     Test reading with EDF_BinaryFileSize bigger than required
                      => currently an unnecessary info is given
        """
        avglist=[9584.23, 9592.64, 9591.69, 9599.7, 9602.51, 9604.29, 9610.97, 9609.86, 9614.14, 9610.52, 9603.12, 9603.27, 9600.22, 9606.86, 9605.26, 9601.37, 9606.09, 9604.51, 9604.45, 9617.5]
        filename = os.path.join(self.root,"02_multi_raw_bf_gblk/rh28a_saxs_00022_raw_binned.ehf")
        test_00(self,filename,avglist)

    def test_edfmulti_raw_dark_raw_bf_gblk(self):
        """
        Linking data of several frames to different parts of a single external
        binary file.
        rh28a_saxs_00003_dark_binned.ehf:
          WARNING:fabio.edfimage:Under-short header frame 2: only 311 bytes
             => this should disappear when reading EDF_BlockBoundary
          INFO:fabio.edfimage:Data stream is padded : 29184 > required 28800 bytes
            => unnecessary, it cannot be avoided, e.g. for saving byte arrays of odd length
        rh28a_saxs_00003_raw_binned.ehf:
          INFO:fabio.edfimage:Data stream is padded : 29184 > required 28800 bytes
            => unnecessary, feature, e.g. for saving byte arrays of odd length
        """
        avglist=[2487.36, 2488.28, 2488.11]
        filename = os.path.join(self.root,"03_multi_raw_dark_bf_gblk/rh28a_saxs_00003_dark_binned.ehf")
        test_00(self,filename,avglist)
        avglist=[9651.25]
        filename = os.path.join(self.root,"03_multi_raw_dark_bf_gblk/rh28a_saxs_00003_raw_binned.ehf")
        test_00(self,filename,avglist)

    def test_edfsingle_raw_bf_gblk_gz(self):
        """
        Test reading gzipped data files linking to an external binary data file.
        Check that the extension .gz is added to the name of the external binary
        data file if it cannot be opened with the binary file name found in the
        header. Both files are gzipped.
        """
        avglist=[25743.2]
        filename = os.path.join(self.root,"04_single_raw_bf_gblk_gz/pj19_frelon_00028_raw.ehf.gz")
        test_00(self,filename,avglist)

    def test_edfsingle_raw_bf_gblk_gz(self):
        """
        Test reading gzipped data files linking to an external binary data file.
        Check that the extension .gz is added to the name of the external binary
        data file if it cannot be opened with the binary file name found in the
        header. The binary data file is not compressed.
        """
        avglist=[25743.2]
        filename = os.path.join(self.root,"05_single_raw_bf_gblk_gz/pj19_frelon_00028_raw.ehf.gz")
        test_00(self,filename,avglist)

    def test_edf6_single_raw_bf_gblk_gz(self):
        """
        Test reading gzipped data files linking to an external binary data file.
        Check that the extension .gz is added to the name of the external binary
        data file if it cannot be opened with the binary file name found in the
        header. Only the binary data file is gzipped.
        """
        avglist=[25743.2]
        filename = os.path.join(self.root,"06_single_raw_bf_gblk_gz/pj19_frelon_00028_raw.ehf")
        test_00(self,filename,avglist)

    def test_edfmulti_raw_bf_gblk_gz(self):
        """
        Test reading the files of the test test_edfmulti_raw_bf_gblk after
        gzipping.
        Test reading from multi frame files, like test_edfmulti_raw_bf_gblk, but
        but with all files gzipped.
        frames 0..16: Test reading from a multi frame edf file with links to
                      external binary files (ehf)
        frame 17:     Test reading without EDF_BinaryFileSize => should be
                      estimated from data size
        frame 18:     Test reading with EDF_BinaryFileSize bigger than required
                      => currently an unnecessary info is given
        frame 19:     Test reading with a wrong EDF_BinaryFileSize that excceeds
                      the real file size. => data must only be read to the end
                      of the binary data file.
        """
        avglist=[9584.23, 9592.64, 9591.69, 9599.7, 9602.51, 9604.29, 9610.97, 9609.86, 9614.14, 9610.52, 9603.12, 9603.27, 9600.22, 9606.86, 9605.26, 9601.37, 9606.09, 9604.51, 9604.45, 9617.5]
        filename = os.path.join(self.root,"07_multi_raw_bf_gblk_gz/rh28a_saxs_00022_raw_binned.ehf.gz")
        test_00(self,filename,avglist)

    def test_pitfalls(self):
        """
        multi5+headerblob_edf1.edf.gz
        multi5+headerblob+headerendinheader_edf1.edf.gz
        multi5+pitfalls_edf1.edf.gz
        multi5+headerblob_edf0.edf.gz
        multi5+headerblob+headerendinheader_edf0.edf.gz
        multi5+pitfalls_edf0.edf.gz
        """

        # edf1 file => must always work
        avglist=[0,1,2,3,4]
        filename = os.path.join(self.root,"08_pitfalls/multi5+pitfalls_edf1.edf.gz")
        test_00(self,filename,avglist)

        # edf1 file with header end pattern in binary blob => must always work
        avglist=[18312,18312,18312,18312,18312]
        filename = os.path.join(self.root,"08_pitfalls/multi5+headerblob_edf1.edf.gz")
        test_00(self,filename,avglist)

        # edf1 file with header end pattern in Title value and binary blob
        # => must always work
        filename = os.path.join(self.root,"08_pitfalls/multi5+headerblob+headerendinheader_edf1.edf.gz")
        test_00(self,filename,avglist)

        # edf0 file => must always work
        avglist=[0,1,2,3,4]
        filename = os.path.join(self.root,"08_pitfalls/multi5+pitfalls_edf0.edf.gz")
        test_00(self,filename,avglist)

        # edf0 file with header end pattern in binary blob => must always work
        avglist=[18312,18312,18312,18312,18312]
        filename = os.path.join(self.root,"08_pitfalls/multi5+headerblob_edf0.edf.gz")
        test_00(self,filename,avglist)

        # edf0 file with header end pattern in Title value and binary blob
        # => usually fails, because the header end pattern is searched
        # by parsing each byte of the header.
        # The next test usually fails, because the value of the header key
        # "Title" contains a header end pattern.
        # There is no safe way of reading such files.
        # When writing the files it must be checked that no special
        # characters are written to header values, especially
        # not '{', '}', ';'.
        # Searching the end_marker in a header in steps of BLOCKSIZE
        # instead of searching through the whole header byte for
        # byte would allow reading most of such files, but success is not
        # guaranteed. => Do not write special character to the header
        # test_00(self,"08_pitfalls/multi5+headerblob+headerendinheader_edf0.edf.gz",avglist)

    def test_special(self):
        """
        09_special/face_ok.edf.gz
        09_special/face_headerendatboundary1.edf.gz
        09_special/face_headerendatboundary2.edf.gz
        09_special/face_headerendatboundary3.edf.gz
        09_special/face_tooshort.edf.gz
        """

        avglist=[0.178897]
        filename = os.path.join(self.root,"09_special/face_ok.edf.gz")
        test_00(self,filename,avglist)

        # The next 3 tests check that edf files can be
        # read where the header end pattern spreads over a
        # BLOCKSIZE boundary. This is usually not a problem
        # for header blocks that are padded to multiples of BLOCKSIZE
        # and that are aligned to BLOCKSIZE boundaries.
        # This error can happen if edf header are modified with
        # an editor.
        # The following tests will fail if distributed  header end
        # patterns are not recognized.
        avglist=[0.178897]
        # header end character '}' at BLOCKSIZE-1 followed by '\n' at BLOCKSIZE
        filename = os.path.join(self.root,"09_special/face_headerendatboundary1.edf.gz")
        test_00(self,filename,avglist)
        # first part of header end pattern '}\r' at BLOCKSIZE-2 followed by '\n' at BLOCKSIZE
        filename = os.path.join(self.root,"09_special/face_headerendatboundary2.edf.gz")
        test_00(self,filename,avglist)
        # header end character '}' at BLOCKSIZE-1 followed by '\n' at BLOCKSIZE
        filename = os.path.join(self.root,"09_special/face_headerendatboundary3.edf.gz")
        test_00(self,filename,avglist)

        # Here, the binary blob is too short
        # like 09_special/face_headerendatboundary1.edf.gz, but in addition
        # the binary blob is too short by 1 byte.
        # Currently, an error message is shown and the data
        # is truncated
        # avglist=[0.178897]
        # filename = os.path.join(self.root,"09_special/face_tooshort.edf.gz")
        # test_00(self,filename,avglist)

    # test files 10 and 11 for testing with to be copied and added later

def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(EdfBlockBoundaryCases))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Wrapper module for the `gzip` library.

Feed this module using a local copy of `gzip` if it exists.
Else it expect to have an available `gzip` library installed in
the Python path.

It should be used like that:

.. code-block::

    from fabio.third_party import gzip

"""

from __future__ import absolute_import

__authors__ = ["Valentin Valls"]
__license__ = "MIT"
__date__ = "28/07/2017"

import sys as __sys

if __sys.version_info < (2, 7):
    # Try to import our local version of six
    from ._local.gzip import *  # noqa
else:
    # Else try to import it from the python path

    # Importing star here is not working
    # from gzip import *  # noqa

    import gzip as __gzip
    for k, v in __gzip.__dict__.items():
        if k.startswith("_"):
            continue
        vars()[k] = v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#

"""Test Esperanto images
"""

from __future__ import print_function, with_statement, division, absolute_import

import fabio.esperantoimage
from ..utilstest import UtilsTest

import unittest
import logging

logger = logging.getLogger(__name__)


class TestEsperanto(unittest.TestCase):
    # filename dim1 dim2 min max mean stddev
    TESTIMAGES = [
        ("sucrose_1s__1_1.esperanto.bz2", 2048, 2048, 0, 65535, 8546.6414, 1500.4198)
    ]

    def test_read(self):
        """
        Test the reading of Mar345 images
        """
        for params in self.TESTIMAGES:
            name = params[0]
            logger.debug("Processing: %s" % name)
            dim1, dim2 = params[1:3]
            shape = dim2, dim1
            mini, maxi, mean, stddev = params[3:]
            obj = fabio.esperantoimage.EsperantoImage()
            obj.read(UtilsTest.getimage(name))

            # self.assertAlmostEqual(mini, obj.getmin(), 2, "getmin [%s,%s]" % (mini, obj.getmin()))
            # self.assertAlmostEqual(maxi, obj.getmax(), 2, "getmax [%s,%s]" % (maxi, obj.getmax()))
            # self.assertAlmostEqual(mean, obj.getmean(), 2, "getmean [%s,%s]" % (mean, obj.getmean()))
            # self.assertAlmostEqual(stddev, obj.getstddev(), 2, "getstddev [%s,%s]" % (stddev, obj.getstddev()))

            self.assertEqual(shape, obj.shape, "dim1")

    def test_header(self):
        for params in self.TESTIMAGES:
            name = params[0]
            logger.debug("Processing: %s" % name)
            obj = fabio.esperantoimage.EsperantoImage()
            obj.read(UtilsTest.getimage(name))

            # The key order is not the same depending on Python2 or 3
            expected_keys = set([
                'IMAGE',
                'SPECIAL_CCD_1',
                'SPECIAL_CCD_2',
                'SPECIAL_CCD_3',
                'SPECIAL_CCD_4',
                'SPECIAL_CCD_5',
                'TIME',
                'MONITOR',
                'PIXELSIZE',
                'TIMESTAMP',
                'GRIDPATTERN',
                'STARTANGLESINDEG',
                'ENDANGLESINDEG',
                'GONIOMODEL_1',
                'GONIOMODEL_2',
                'WAVELENGTH',
                'MONOCHROMATOR',
                'ABSTORUN',
                'HISTORY',
                'ESPERANTO_FORMAT'])

            self.assertEqual(set(obj.header.keys()), expected_keys)


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestEsperanto))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



from io import BytesIO
from struct import pack, unpack

import numpy as np

unpack_ = unpack
unpack = lambda fmt, buff: unpack_(fmt, buff)[0]


def compress(frame):
    """compress a frame using the agi_bitfield algorithm
    :param frame: numpy.ndarray
    :returns bytes
    """

    dim = frame.shape
    buffer = BytesIO()

    row_adresses = BytesIO()

    for frame_row in range(0, dim[0]):
        row_adresses.write(pack("I", buffer.tell() + 4))
        compress_row(frame[frame_row], buffer)

    data_size = pack("I", buffer.tell())
    buffer.write(row_adresses.getvalue())

    row_adresses.close()
    return data_size + buffer.getvalue()


def compress_row(data, buffer):
    """compress a single row
    :arg data numpy.array
    :arg buffer io.BytesIO
    """

    first_pixel = data[0]
    pixel_diff = data[1:] - data[:-1]

    write_escaped(first_pixel, buffer)

    n_fields = len(pixel_diff) // 16
    n_restpx = len(pixel_diff) % 16

    for field in range(0, n_fields):
        fielda = pixel_diff[:8]
        fieldb = pixel_diff[8:16]

        len_a, len_b = get_fieldsize(fielda), get_fieldsize(fieldb)
        len_byte = (len_b << 4) | (0xf & len_a)
        buffer.write(pack("B", len_byte))

        of_buff = BytesIO()
        compressed_fielda = compress_field(fielda, len_a, of_buff)
        compressed_fieldb = compress_field(fieldb, len_b, of_buff)

        buffer.write(compressed_fielda + compressed_fieldb + of_buff.getvalue())

        pixel_diff = pixel_diff[16:]

    for restpx in range(0, n_restpx):
        write_escaped(pixel_diff[restpx], buffer)


def decompress(comp_frame, dimensions):
    """decompresses a frame that was compressed using the agi_bitfield algorithm
    :param comp_frame: bytes
    :param dimensions: tuple
    :return numpy.ndarray
    """

    row_count, col_count = dimensions

    # read data components (row indices are ignored)
    data_size = unpack("I", comp_frame[:4])
    data_block = BytesIO(comp_frame[4:])

    output = []

    # iterate over rows
    for row_index in range(0, row_count):
        output.append(decompress_row(data_block, col_count))

    return np.array(output, dtype="int32").cumsum(axis=1)


def decompress_row(buffer, row_length):
    """decompress a single row
    :param buffer: io.BytesIO
    :param row_length: int
    :returns list
    """

    first_pixel = read_escaped(buffer)

    n_fields = (row_length - 1) // 16
    n_restpx = (row_length - 1) % 16

    pixels = [first_pixel]

    for field in range(n_fields):
        lb = unpack("B", buffer.read(1))
        len_b, len_a = read_len_byte(lb)

        field_a = decode_field(buffer.read(len_a))
        field_b = decode_field(buffer.read(len_b))

        undo_escapes(field_a, len_a, buffer)
        undo_escapes(field_b, len_b, buffer)

        pixels += field_a
        pixels += field_b

    for restpx in range(n_restpx):
        pixels.append(read_escaped(buffer))

    return pixels


def get_fieldsize(array):
    """determine the fieldsize to store the given values
    :param array numpy.array
    :returns int
    """

    def fieldsize(n):
        return min(8, max(1, abs(n.item()).bit_length() + 1))

    return max(fieldsize(array.max()), fieldsize(array.min()))


def compress_field(ifield, fieldsize, overflow_table):
    """compress a field with given size
    :param ifield: numpy.ndarray
    :param fieldsize: int
    :param overflow_table: io.BytesIO
    :returns int
    """

    compressed_field = np.uint64(0)  # uint64
    fieldsize = np.uint8(fieldsize)

    mask_ = mask(fieldsize)
    conv_ = conv(fieldsize)
    for i, elem in enumerate(ifield):
        if -127 <= elem < 127:
            val = np.uint8((elem + conv_) & mask_)
        elif -32767 <= elem < 32767:
            val = np.uint8(254)
            overflow_table.write(pack("h", elem))
        else:
            val = np.uint8(255)
            overflow_table.write(pack("i", elem))
        compressed_field = val | (compressed_field << fieldsize)
    return pack("Q", compressed_field)[:fieldsize]


def decode_field(field):
    """decodes a field from bytes
    :param field: bytes
    :returns list
    """
    size = len(field)
    assert 0 < size <= 8, str(size)
    field = unpack("Q", field + (b'\x00' * (8 - size)))
    mask_ = mask(size)
    values = []
    for i in range(0, 8):
        values.append((field >> (size * (7 - i)) & mask_))
    return values


def read_len_byte(lb):
    """parses the length byte and returns the sizes of the next twofields
    :param lb: int/byte
    :returns tuple
    """
    return lb >> 4,  lb & 0xf


def mask(length):
    """returns a bit mask of the given length
    :param length: int
    :returns int
    """
    return (1 << length) - 1


def conv(length):
    """returns the conversion value of the given length
    :param length: int
    :returns int
    """
    return (1 << length - 1) - 1


def write_escaped(value, buffer):
    """write an value to the buffer and escape when overflowing one byte
    :param value: int
    :param buffer: io.BytesIO
    """
    if -127 <= value < 127:
        buffer.write(pack("B", value + 127))
    elif -32767 < value < 32767:
        buffer.write(b'\xfe' + pack("h", value))
    else:
        buffer.write(b'\xff' + pack("i", value))


def read_escaped(buffer):
    """reads byte value that might be escaped from a buffer
    :param buffer: io.BytesIO
    :returns int
    """
    byte = buffer.read(1)
    if byte == b'\xfe':  # two byte overflow
        return unpack("h", buffer.read(2))
    elif byte == b'\xff':  # four byte overflow
        return unpack("i", buffer.read(4))
    else:  # no overflow
        return unpack("B", byte) - 127


def undo_escapes(field, length, buffer):
    """undo escaping of values in a field
    :param field: list
    :param length: int
    :param buffer: io.BytesIO
    """
    conv_ = conv(length)
    for i, val in enumerate(field):
        if val == 0xfe:
            field[i] = unpack("h", buffer.read(2))
        elif val == 0xff:
            field[i] = unpack("i", buffer.read(4))
        else:
            field[i] = field[i] - conv_



from __future__ import print_function, with_statement, division, absolute_import

import io
import logging
import unittest
from struct import pack

import numpy as np

from fabio.compression import agi_bitfield

logger = logging.getLogger(__name__)


class TestUtil(unittest.TestCase):
    def test_get_fieldsize(self):
        test_cases = [
            (np.zeros(8, dtype="int32"), 1),
            (np.array([1024] * 8, dtype="int32"), 8),
            (np.arange(0, 7, dtype="int32"), 4),
            (np.array([-1, -2, 2, 1] * 2, dtype="int32"), 3)
        ]
        for array, result in test_cases:
            fs = agi_bitfield.get_fieldsize(array)
            self.assertEqual(result, fs)

    def test_compress_decode_field_overflow(self):
        data = np.array([1, 2, 3, 4, 255, 40000, -4, 2], dtype="int32")
        fs = agi_bitfield.get_fieldsize(data)

        oft = io.BytesIO()
        cf = agi_bitfield.compress_field(data, fs, oft)
        self.assertEqual(pack("h", 255) + pack("i", 40000), oft.getvalue())

        mask_ = agi_bitfield.mask(fs)
        dec = agi_bitfield.decode_field(cf)
        self.assertEqual([1 + 127, 2 + 127, 3 + 127, 4 + 127, 0xfe & mask_, 0xff & mask_, -4 + 127, 2 + 127], dec)

    def test_compress_decompress_field_small(self):
        data = np.array([1, 0] * 4, dtype="int32")
        fs = agi_bitfield.get_fieldsize(data)

        oft = io.BytesIO()
        cf = agi_bitfield.compress_field(data, fs, oft)
        self.assertEqual(b'', oft.getvalue())

        conv_ = agi_bitfield.conv(fs)
        dec = agi_bitfield.decode_field(cf[:fs])
        self.assertEqual([1 + conv_, 0 + conv_] * 4, dec)

    def test_full_compress_decompress_field(self):
        data = np.array([1, 2, 3, 4, 255, -40000, -4, 2], dtype="int32")
        fs = agi_bitfield.get_fieldsize(data)

        oft = io.BytesIO()
        cf = agi_bitfield.compress_field(data, fs, oft)

        oft = io.BytesIO(oft.getvalue())  # do this to simulate actual conditions. the data is passed between
        # encoding and decoding as bytes so the index would not be preserved

        field = agi_bitfield.decode_field(cf[:fs])
        agi_bitfield.undo_escapes(field, fs, oft)
        self.assertEqual(list(data), field)

    def test_mask_conv(self):
        mask_cases = [
            (1, 1),
            (4, 0b1111),
            (8, 0xff)
        ]
        for len, msk in mask_cases:
            self.assertEqual(msk, agi_bitfield.mask(len))
        conv_cases = [
            (1, 0),
            (4, 0b111),
            (8, 0b1111111)
        ]
        for len, cnv in conv_cases:
            self.assertEqual(cnv, agi_bitfield.conv(len))


class TestRow(unittest.TestCase):
    def test_compress_row_zero(self):
        data = np.zeros(256, dtype="int32")
        buffer = io.BytesIO()

        agi_bitfield.compress_row(data, buffer)

        buffer.seek(0)

        decompressed = agi_bitfield.decompress_row(buffer, 256)
        decompressed = np.array(decompressed, dtype="int32").cumsum()
        self.assertTrue(np.array_equal(decompressed, data))

    def test_compress_row_small(self):
        data = np.random.randint(0, 5, 256, dtype="int32")
        buffer = io.BytesIO()

        agi_bitfield.compress_row(data, buffer)

        buffer.seek(0)

        decompressed = agi_bitfield.decompress_row(buffer, 256)
        decompressed = np.array(decompressed, dtype="int32").cumsum()
        self.assertTrue(np.array_equal(decompressed, data))

    def compress_row_overflows(self):
        data = np.random.randint(-255, 255, 256, dtype="int32")
        buffer = io.BytesIO()

        agi_bitfield.compress_row(data, buffer)

        buffer.seek(0)

        decompressed = agi_bitfield.decompress_row(buffer, 256)
        decompressed = np.array(decompressed, dtype="int32").cumsum()
        self.assertTrue(np.array_equal(decompressed, data))


class TestCompression(unittest.TestCase):
    def test_full_rand(self):
        data = np.random.randint(-255, 255, (256, 256))
        compressed = agi_bitfield.compress(data)
        uncompressed = agi_bitfield.decompress(compressed, data.shape)
        self.assertTrue(np.array_equal(data, uncompressed))


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestUtil))
    testsuite.addTest(loadTests(TestRow))
    testsuite.addTest(loadTests(TestCompression))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())
